import textwrap
import time
import httpx
import asyncio
import atexit
import signal
import json
import threading
from typing import Optional, Any, List, Dict, Sequence

import docker
from docker.models.containers import Container
from pydantic import BaseModel

from codearkt.llm import ChatMessage
from codearkt.tools import fetch_tools


IMAGE = "phoenix120/codearkt_http"
MEM_LIMIT = "512m"
CPU_QUOTA = 50000
CPU_PERIOD = 100000
EXEC_TIMEOUT = 600
PIDS_LIMIT = 64
CLIENT = None
NET_NAME = "sandbox_net"


class ExecResult(BaseModel):  # type: ignore
    stdout: str
    error: str | None = None
    result: Any | None = None

    def to_messages(self, tool_call_id: str) -> List[ChatMessage]:
        image_content: List[Dict[str, Any]] | None = None
        output: str = "Stdout:\n" + self.stdout + "\n\n"
        if self.result is not None:
            try:
                json_result = json.loads(str(self.result))
                if isinstance(json_result, dict) and "image_base64" in json_result:
                    image_base64 = json_result["image_base64"]
                    image_content = [
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:image/jpeg;base64,{image_base64}"},
                        }
                    ]
            except json.JSONDecodeError:
                pass
            if not image_content:
                output += "Last expression:\n" + str(self.result) + "\n\n"
        if self.error:
            output += "Error: " + self.error
        output = output.strip()

        messages = []
        messages.append(ChatMessage(role="tool", content=output, tool_call_id=tool_call_id))
        if image_content:
            messages.append(ChatMessage(role="user", content=image_content))
        return messages


def init_docker() -> docker.DockerClient:
    client = docker.from_env()
    try:
        client.ping()  # type: ignore
    except docker.errors.DockerException as exc:
        raise RuntimeError(
            "Docker daemon is not running or not accessible – skipping PythonExecutor setup."
        ) from exc

    try:
        client.images.get(IMAGE)
    except docker.errors.ImageNotFound:
        try:
            client.images.pull(IMAGE)
        except docker.errors.DockerException as exc:
            raise RuntimeError(
                f"Docker image '{IMAGE}' not found locally and failed to pull automatically."
            ) from exc
    except docker.errors.DockerException as exc:
        raise RuntimeError("Failed to query Docker images – ensure Docker is available.") from exc
    return client


class PythonExecutor:
    def __init__(
        self,
        tool_names: Sequence[str] = tuple(),
        session_id: Optional[str] = None,
        mcp_server_port: int = 5055,
    ) -> None:
        global CLIENT
        if not CLIENT:
            CLIENT = init_docker()
        client = CLIENT
        try:
            net = client.networks.get(NET_NAME)
        except docker.errors.NotFound:
            net = client.networks.create(
                NET_NAME,
                driver="bridge",
                internal=False,
                options={
                    "com.docker.network.bridge.enable_ip_masquerade": "false",
                },
            )

        self.session_id = session_id
        self.tool_names = tool_names
        self.mcp_server_port = mcp_server_port
        self.tools_are_checked = False

        self.container: Optional[Container] = client.containers.run(
            IMAGE,
            detach=True,
            auto_remove=True,
            ports={"8000/tcp": None},
            mem_limit=MEM_LIMIT,
            cpu_period=CPU_PERIOD,
            cpu_quota=CPU_QUOTA,
            pids_limit=PIDS_LIMIT,
            cap_drop=["ALL"],
            read_only=True,
            tmpfs={"/tmp": "rw,size=64m", "/run": "rw,size=16m"},
            security_opt=["no-new-privileges"],
            extra_hosts={"host.docker.internal": "host-gateway"},
            sysctls={"net.ipv4.ip_forward": "0"},
            network=net.name,
            dns=[],
            environment={"SERVER_PORT": str(mcp_server_port)},
        )
        self.url = self._get_url()

        atexit.register(self.cleanup)

        if threading.current_thread() is threading.main_thread():
            self._attach_signal_handler(signal.SIGINT)
            self._attach_signal_handler(signal.SIGTERM)

    def _attach_signal_handler(self, sig: int) -> None:
        previous = signal.getsignal(sig)

        def _handler(signum: int, frame: Optional[Any] = None) -> None:
            try:
                self.cleanup()
            finally:
                if callable(previous):
                    previous(signum, frame)

        signal.signal(sig, _handler)

    async def invoke(self, code: str) -> ExecResult:
        await self._check_tools()
        await self._wait_for_ready()
        result = await self._call_exec(code)
        return result

    async def _check_tools(self) -> None:
        if not self.tool_names or self.tools_are_checked:
            return
        available_tools = await fetch_tools(f"http://localhost:{self.mcp_server_port}")
        available_tool_names = [tool.name for tool in available_tools]
        for tool_name in self.tool_names:
            if tool_name not in available_tool_names:
                raise ValueError(f"Tool {tool_name} not found in MCP server")
        self.tools_are_checked = True

    async def _call_exec(self, code: str) -> ExecResult:
        payload = {
            "code": textwrap.dedent(code),
            "session_id": self.session_id,
            "tool_names": self.tool_names,
        }

        async with httpx.AsyncClient(limits=httpx.Limits(keepalive_expiry=0)) as client:
            # TODO: cancel internal requests after timeout
            # TODO: proper error for timeout
            resp = await client.post(f"{self.url}/exec", json=payload, timeout=EXEC_TIMEOUT)
            resp.raise_for_status()
            out = resp.json()
            result: ExecResult = ExecResult.model_validate(out)
            return result

    def cleanup(self) -> None:
        if self.container is None:
            return
        try:
            self.container.remove(force=True)
        except Exception:
            pass
        finally:
            self.container = None

    def _get_url(self) -> str:
        assert self.container is not None
        self.container.reload()
        ports = self.container.attrs["NetworkSettings"]["Ports"]
        mapping = ports["8000/tcp"][0]
        return f"http://localhost:{mapping['HostPort']}"

    async def _wait_for_ready(self, max_wait: int = 60) -> None:
        start_time = time.time()
        while time.time() - start_time < max_wait:
            try:
                output = await self._call_exec("print('ready')")
                assert output.stdout.strip() == "ready"
                return
            except (httpx.RequestError, httpx.TimeoutException, AssertionError):
                pass
            await asyncio.sleep(0.1)
        raise RuntimeError("Container failed to become ready within timeout")
